// David Pilling
// log Oregon sensor readings
//


#include "PinChangeInt.h"
#include "LowPower.h"



#include <SerialCommand.h>

#include <Time.h>

#include <EEPROM.h>



#define pin_led 13
#define pin_clock_power 12
#define pin_battery_voltage A0
#define pin_sense_power 9
#define pin_battery_charge 8


#define MAX_SENSORS 8

typedef struct bootblockstr
{
 time_t        basetime;
 unsigned long period; 
 int           recordsize;
 int           memorysize;
 int           nsensors;
 int           sensormodel[MAX_SENSORS];
 byte          sensorchannel[MAX_SENSORS]; 

 unsigned long lastgoodvalue;
           int targetmiss;
 unsigned int  blank1;
 unsigned int  blank2;


} bootblockstr;


int runingoffusb;

bootblockstr bootblock;



unsigned long batterytime=0;
int  batterycharging=0;
int  batteryruningoffusb;
int  batterycount;
long batteryaverage10;
long batteryaverage20;
int  batteryload10;



int loopstep=1;
int loopphase=0;

char pattern1[8]={1,1,1,1,0,0,0,0};
char pattern2[8]={1,0,0,1,0,0,0,0};
char * pattern=pattern1;

void ledupdate(void)
{
 static long updatetime=0;
 long nowtime=millis();
  
 if((nowtime-updatetime)>200)
 {
  digitalWrite(pin_led, pattern[loopphase]);   // turn the LED on (HIGH is the voltage level)
  loopphase++;
  if(loopphase>7) loopphase=0;
  updatetime=nowtime;
 }
}



void writelastgoodvalue(void)
{
 int    i;
 char * p;
 char * q;

 p=(char*)&bootblock.lastgoodvalue;
 q=(char*)&bootblock;
 for(i=0;i<(sizeof(bootblock.lastgoodvalue)+sizeof(bootblock.targetmiss)+sizeof(bootblock.blank1)+sizeof(bootblock.blank2));i++) 
 {
  EEPROM.write(p-q,*p++);
 }    
}  


void writebootblock(void)
{
 int    i;
 char * p;

 p=(char*)&bootblock;
 for(i=0;i<sizeof(bootblock);i++) 
 {
  EEPROM.write(i,*p++);
 }  
}




void readbootblock(void)
{
 int    i;
 char * p;

 p=(char*)&bootblock;
 for(i=0;i<sizeof(bootblock);i++) 
 {
  *p++=EEPROM.read(i);
 }  
}  





// https://code.google.com/p/tinkerit/wiki/SecretVoltmeter

long readVcc() {
  long result;
  // Read 1.1V reference against AVcc
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = 1126400L / result; // Back-calculate AVcc in mV
  return result;
}








void power_up(void)
{
 powerup_clock(); 
}


void power_down(void)
{
 Serial.flush(); 
 powerdown_clock(); 
}  



//*** Process serial port commands

SerialCommand sCmd;



void cmdunrecognized(const char *command)
{
 Serial.println("What?"); 
}  


void cmdreadtime(void)
{
 ReadDS3231(); 
}


void cmdreset(void)
{
 char buffer[32];
 
 bootblock.basetime=nowtime();

 bootblock.nsensors=4;
 bootblock.sensormodel[0]=0x1A2D;
 bootblock.sensormodel[1]=0xEA4C;
 bootblock.sensormodel[2]=0x1ABC;
 bootblock.sensormodel[3]=0xEA4C;
 bootblock.sensorchannel[0]=2;
 bootblock.sensorchannel[1]=4;
 bootblock.sensorchannel[2]=2;
 bootblock.sensorchannel[3]=1;

 bootblock.recordsize=packsensors(buffer,sizeof(buffer));
 bootblock.period=1800L;
 bootblock.memorysize=4096;
 bootblock.lastgoodvalue=0;

 writebootblock();
 
 clock_reset();
}


void cmdstatus(void)
{
 Serial.print("basetime:");
 Serial.println(bootblock.basetime);
 
 Serial.print("period:"); 
 Serial.println(bootblock.period); 

 Serial.print("recordsize:");
 Serial.println(bootblock.recordsize);
 
 Serial.print("memorysize:");
 Serial.println(bootblock.memorysize);
 
 Serial.print("lastgoodvalue:");
 Serial.println(bootblock.lastgoodvalue);

 Serial.print("vcc:");
 Serial.println(readVcc());
 
 Serial.print("now:");
 Serial.println(nowtime());
 
 Serial.print("batterycount:");
 Serial.println(batterycount);
 
 Serial.print("batteryaverage10:");
 Serial.println(batteryaverage10);
 
 Serial.print("batteryaverage20:");
 Serial.println(batteryaverage20);
 
 Serial.print("batteryload10:");
 Serial.println(batteryload10);

 Serial.print("batterycharging:");
 Serial.println(batterycharging);

 Serial.print("batteryruningoffusb:");
 Serial.println(batteryruningoffusb);

 Serial.print("targetmiss:");
 Serial.println(bootblock.targetmiss);
 
 Serial.print("blank1:");
 Serial.println(bootblock.blank1);
 
 Serial.print("blank2:");
 Serial.println(bootblock.blank2);
}  

void cmddebug(void)
{
  
 bootblock.targetmiss=0;
 bootblock.blank1=0;
 bootblock.blank2=0;
 writelastgoodvalue();
 
 cmdstatus(); 
}


void cmdreadvalue()
{
 char * arg;
 unsigned long pin;
 unsigned int location;
 byte buffer[32];
 int i;
 
 arg = sCmd.next();
 if(arg!=NULL)
 {
  pin=atol(arg);

  location=(((unsigned long)bootblock.recordsize)*pin) % ((unsigned long)bootblock.memorysize);

  i2c_eeprom_read_buffer(0x57,location,buffer,bootblock.recordsize);
 
  Serial.print("rv:");
  for(i=0;i<bootblock.recordsize;i++)
  {
   Serial.print(itox(buffer[i]>>4));
   Serial.print(itox(buffer[i]&0xF));
  }
  Serial.print(":");
  Serial.print(pin);
  Serial.println();  
    
 }
}


void cmdpower()
{
 char * arg;
 unsigned int pin;
 
 arg = sCmd.next();
 if(arg!=NULL)
 {
  pin=atoi(arg);
  if(pin==0) power_down();
  else       power_up();
    
 }
}



void cmdreadpin()
{
 char * arg;
 int    rpin; 
 
 arg = sCmd.next();
 if(arg!=NULL)
 {
  rpin=atoi(arg);
  
  Serial.print("rd "); 
  Serial.println(digitalRead(rpin)); 
 }
}  


void cmdwritepin()
{
 char * arg;
 int    wpin; 
 int    val;
 
 arg = sCmd.next();
 if(arg!=NULL)
 {
  wpin=atoi(arg);

  arg = sCmd.next();
  if(arg!=NULL)
  {
   val=atoi(arg);
   
   digitalWrite(wpin,val); 
  } 
 }
}  


void cmdreadanalogpin()
{
 char * arg;
 int    rpin; 
 long   vcc;
 
 arg = sCmd.next();
 if(arg!=NULL)
 {
  rpin=atoi(arg);

  int sensorValue = analogRead(A0+rpin);
  vcc=readVcc();  
  float voltage = vcc-sensorValue * (vcc / 1023.0);
  
  Serial.print("ra "); 
  Serial.print(sensorValue);
  Serial.print(" ");
  Serial.println(voltage); 
 }
}  



void setup_serial(void)
{
 Serial.begin(9600);

 sCmd.addCommand("rt",cmdreadtime);
 sCmd.addCommand("ws",cmdwriteseconds);
 sCmd.addCommand("wm",cmdwriteminutes);
 sCmd.addCommand("wh",cmdwritehours);
 sCmd.addCommand("wd",cmdwritedate); 
 sCmd.addCommand("wn",cmdwritemonth); 
 sCmd.addCommand("wy",cmdwriteyear); 
 sCmd.addCommand("rx",RXdata); 
 sCmd.addCommand("xx",cmdreset); 
 sCmd.addCommand("st",cmdstatus);
 sCmd.addCommand("rv",cmdreadvalue);
 sCmd.addCommand("pw",cmdpower); 
 sCmd.addCommand("rd",cmdreadpin); 
 sCmd.addCommand("wr",cmdwritepin); 
 sCmd.addCommand("ra",cmdreadanalogpin);  
 sCmd.addCommand("db",cmddebug);  
  
 sCmd.setDefaultHandler(cmdunrecognized);  
}


//*** end of serial port commands

int senseflag;

void pinsense() 
{
 senseflag=1;
}


void pinsensehandler(void)
{
 if(senseflag)
 {
//  if(digitalRead(pin_sense_power)) pattern=pattern2;
//  else pattern=pattern1;
  
  if(digitalRead(pin_sense_power)) runingoffusb=0;
  else runingoffusb=1;
  
  senseflag=0;
  
  
 } 
}  


void pinsensesetup(void)
{
  
 pinMode(pin_sense_power, INPUT); 
 pinMode(pin_led, OUTPUT); 

 PCintPort::attachInterrupt(pin_sense_power,&pinsense, CHANGE);

 if(digitalRead(pin_sense_power)) pattern=pattern2;
 else pattern=pattern1;
 
 if(digitalRead(pin_sense_power)) runingoffusb=0;
 else runingoffusb=1;

}


void analogsetup(void)
{
// analogReference(INTERNAL);

 DIDR0=0x1;
 
}





void battery_setup(void)
{
 pinMode(pin_battery_charge, OUTPUT); 
 digitalWrite(pin_battery_charge,0);
 batteryload10=5000;

}


void battery_read(void)
{
 long vcc; 

 int sensorValue = analogRead(pin_battery_voltage);
 vcc=readVcc();  
 long voltage = vcc-(sensorValue * vcc) / 1023L;
 if(batterycount==0) batteryaverage10=batteryaverage20=voltage; 
 else
 {
  batteryaverage10=(9*batteryaverage10+voltage)/10;
  batteryaverage20=(19*batteryaverage20+voltage)/20;
 }  


   Serial.print("bloop count ");
   Serial.print(batterycount);
   Serial.print(" sensor value ");
   Serial.print(sensorValue);
   Serial.print(" 10=");
   Serial.print(batteryaverage10);
   if(voltage<batteryaverage10) Serial.print("*"); else Serial.print(" ");
   Serial.print(" 20=");   
   Serial.print(batteryaverage20);
   if(voltage<batteryaverage20) Serial.print("*"); else Serial.print(" ");
   Serial.print(" now=");   
   Serial.println(voltage); 
  
}  

void battery_loop(void)
{
 unsigned long nowtime=millis();
 long        vcc;

 if((nowtime-batterytime)>=60000)
 {
  if(runingoffusb)
  { 
   if(!batteryruningoffusb)
   {
    batteryruningoffusb=1;

    if(batterycount>10) 
    {
     batteryload10=batteryaverage10;
    }
    else
    {
     batteryload10=5000; 
    }  
    batterycount=0;      
   }  

   battery_read();
 
   if(batterycount==10 && (batteryaverage10<3900 || batteryload10<3900)) 
   {
    digitalWrite(pin_battery_charge,1);
    batterycharging=1;
    Serial.println("Battery starts to charge"); 
   }
  
   // if(batteryaverage10==batteryaverage20)
   if((batteryaverage10<batteryaverage20) && batterycharging)
   {
    digitalWrite(pin_battery_charge,0);
    batterycharging=0;
    Serial.println("Battery charge ends");     
   } 
    
   batterycount++;
  }
  else
  {
   if(batteryruningoffusb)
   {
    batteryruningoffusb=0;
    if(batterycharging)
    {
     digitalWrite(pin_battery_charge,0);
     batterycharging=0;
    } 
    batterycount=0;
   }
   battery_read();
   batterycount++;   
  }  
  batterytime=nowtime;
 } 
}  





void setup () 
{
 readbootblock(); 
 runingoffusb=1;

 analogsetup();
 pinsensesetup();
 
 power_up();
 setup_serial();
 clock_setup();
 battery_setup(); // after pinsensesetup
 RXsetup();
 

}



void loop () 
{
 if(runingoffusb)
 {
  if(Serial.available()>0) 
  {
   sCmd.readSerial();
  }
  RXloop();
//  ledupdate();
  pinsensehandler();
  clock_loop();
  battery_loop();
  
//  clock_loop();
 }
 else
 {
  clock_loop(); // check for interrupts 
  battery_loop();
  if(clock_sleep_ok(10))
  {
   delay(1000);
   power_down(); 
   delay(1000);
//    Serial.print("power down2");
   LowPower.powerDown(SLEEP_FOREVER, ADC_OFF, BOD_OFF);
//  delay(5000);
   power_up();
   delay(1000);
   pinsensehandler(); // put it here to break out of loop if USB power is up
  } 
 } 
}




